OBJDIR	:= obj

# Cross-compiler toolchain
CC	:= $(GCCPREFIX)gcc -pipe
CXX	:= $(GCCPREFIX)c++ -pipe
AS	:= $(GCCPREFIX)as
AR	:= $(GCCPREFIX)ar
LD	:= $(GCCPREFIX)ld
OBJCOPY	:= $(GCCPREFIX)objcopy
OBJDUMP	:= $(GCCPREFIX)objdump
NM	:= $(GCCPREFIX)nm
STRIP	:= $(GCCPREFIX)strip

# Native commands
NCC	:= gcc $(CC_VER) -pipe
TAR	:= tar
PERL	:= perl

# Compiler flags
# Note that -O2 is required for the boot loader to fit within 512 bytes;
# -fno-builtin is required to avoid refs to undefined functions in the kernel.
CFLAGS	:= $(CFLAGS) $(DEFS) $(LABDEFS) -Os -fno-builtin -I. -MD -Wall -Wno-format -Werror -ggdb

# Different flags for kernel and user programs
KERN_CFLAGS := $(CFLAGS) -DWEENYOS_KERNEL
USER_CFLAGS := $(CFLAGS) -DWEENYOS_USER


# This magic automatically generates makefile dependencies
# for header files included from C source files we compile,
# and keeps those dependencies up-to-date every time we recompile.
# See 'mergedep.pl' for more information.
$(OBJDIR)/.deps: $(wildcard $(OBJDIR)/*.d)
	@mkdir -p $(@D)
	@$(PERL) mergedep.pl $@ $^

-include $(OBJDIR)/.deps
-include conf/date.mk


# try to infer the correct GCCPREFIX
conf/gcc.mk:
	@if i386-jos-elf-objdump -i 2>&1 | grep '^elf32-i386$$' >/dev/null 2>&1; \
	then echo 'GCCPREFIX=i386-jos-elf-' >conf/gcc.mk; \
	elif objdump -i 2>&1 | grep '^elf32-i386$$' >/dev/null 2>&1; \
	then echo 'GCCPREFIX=' >conf/gcc.mk; \
	else echo "***" 1>&2; \
	echo "*** Error: Couldn't find an i386-*-elf version of GCC/binutils." 1>&2; \
	echo "*** Is the directory with i386-jos-elf-gcc in your PATH?" 1>&2; \
	echo "*** If your i386-*-elf toolchain is installed with a command" 1>&2; \
	echo "*** prefix other than 'i386-jos-elf-', set your GCCPREFIX" 1>&2; \
	echo "*** environment variable to that prefix and run 'make' again." 1>&2; \
	echo "*** To turn off this error, run 'echo GCCPREFIX= >conf/gcc.mk'." 1>&2; \
	echo "***" 1>&2; exit 1; fi
	@f=`grep GCCPREFIX conf/gcc.mk | sed 's/.*=//'`; if echo $$f | grep '^[12]\.' >/dev/null 2>&1; then echo "***" 1>&2; \
	echo "*** Error: Your gcc compiler is too old." 1>&2; \
	echo "*** The labs will only work with gcc-3.0 or later, and are only" 1>&2; \
	echo "*** tested on gcc-3.3 and later." 1>&2; \
	echo "***" 1>&2; exit 1; fi


# The mkbootdisk program
mkbootdisk: mkbootdisk.c
	@echo + hostcc mkbootdisk.c
	$(V)gcc mkbootdisk.c -o mkbootdisk


# For deleting the build
clean:
	@echo + clean
	$(V)rm -rf $(OBJDIR) *.img mkbootdisk

realclean: clean
	$(V)rm -rf $(DISTDIR)-$(USER).tar.gz $(DISTDIR)-$(USER)

distclean: realclean
	$(V)rm -rf conf/gcc.mk


# Distribute weensyos
DISTDIR = weensyos2

distdir:
	@/bin/rm -rf $(DISTDIR)
	mkdir $(DISTDIR)
	for i in .bochsrc COPYRIGHT GNUmakefile bootstart.S elf.h mergedep.pl mkbootdisk.c schedos-1.c schedos-2.c schedos-3.c schedos-4.c schedos-app.h schedos-boot.c schedos-kern.c schedos-kern.h schedos-symbols.ld schedos-trap.S schedos-x86.c schedos.h types.h x86.h x86struct.h x86sync.h conf conf/rules.mk; do \
	[ -d $$i ] && mkdir $(DISTDIR)/$$i; \
	[ -f $$i ] && ln $$i $(DISTDIR)/$$i; done
	echo >$(DISTDIR)/conf/date.mk "PACKAGEDATE="`date`

dist: distdir
	$(TAR) czf $(DISTDIR).tar.gz $(DISTDIR)
	/bin/rm -rf $(DISTDIR)

tarball: realclean
	@echo + mk $(DISTDIR)-$(USER).tar.gz
	$(V)mkdir $(DISTDIR)-$(USER)
	$(V)tar cf - `ls -a | grep -v '^\.*$$' | grep -v '^$(DISTDIR)*'` | (cd $(DISTDIR)-$(USER) && tar xf -)
	$(V)/bin/rm -rf `find $(DISTDIR)-$(USER) -name CVS -print`
	$(V)date > $(DISTDIR)-$(USER)/tarballstamp
	$(V)tar cf $(DISTDIR)-$(USER).tar $(DISTDIR)-$(USER)
	$(V)gzip $(DISTDIR)-$(USER).tar
	$(V)/bin/rm -rf $(DISTDIR)-$(USER)

handin: tarball
	@echo + handin
	$(V)uuencode $(DISTDIR)-$(USER).tar.gz < $(DISTDIR)-$(USER).tar.gz | mail -s $(DISTDIR)-$(USER).tar.gz osp-s05-handin@pdos.lcs.mit.edu

run-%: %.img
	@echo + bochs $<
	$(V)/bin/mv .bochsrc .bochsrc~
	$(V)sed 's/path="[^"]*"/path="$<"/' <.bochsrc~ >.bochsrc
	$(V)/bin/rm .bochsrc~
	$(V)bochs -q


# Patch targets
# Create a patch from ../$(DISTDIR).tar.gz.
patch-extract-tarball:
	@test -r ../$(DISTDIR).tar.gz || (echo "***" 1>&2; \
	echo "*** Can't find '../$(DISTDIR).tar.gz'.  Download it" 1>&2; \
	echo "*** into my parent directory and try again." 1>&2; \
	echo "***" 1>&2; false)
	@(gzcat ../$(DISTDIR).tar.gz 2>/dev/null || zcat ../$(DISTDIR).tar.gz) | tar xf -

patch-check-date: patch-extract-tarball
	@pkgdate=`grep PACKAGEDATE $(DISTDIR)/conf/date.mk | sed 's/.*=//'`; \
	test "$(PACKAGEDATE)" = "$$pkgdate" || (echo "***" 1>&2; \
	echo "*** The ../$(DISTDIR).tar.gz tarball was created on $$pkgdate," 1>&2; \
	echo "*** but your work directory was expanded from a tarball created" 1>&2; \
	echo "*** on $(PACKAGEDATE)!  I can't tell the difference" 1>&2; \
	echo "*** between your changes and the changes between the tarballs," 1>&2; \
	echo "*** so I won't create an automatic patch." 1>&2; \
	echo "***" 1>&2; false)

patch.diff: patch-extract-tarball always
	@rm -f patch.diff
	@for f in `cd $(DISTDIR) && find . -type f -print`; do \
	if diff -u $(DISTDIR)/"$$f" "$$f" >patch.diffpart || [ "$$f" = ./boot/lab.mk ]; then :; else \
	echo "*** $$f differs; appending to patch.diff" 1>&2; \
	echo diff -u $(DISTDIR)/"$$f" "$$f" >>patch.diff; \
	cat patch.diffpart >>patch.diff; \
	fi; done
	@for f in `find . -name $(DISTDIR) -prune -o -name obj -prune -o -name "patch.diff*" -prune -o -name '*.rej' -prune -o -name '*.orig' -prune -o -type f -print`; do \
	if [ '(' '!' -f $(DISTDIR)/"$$f" ')' -a '(' "$$f" != ./kern/appkernbin.c ')' -a '(' "$$f" != ./conf/gcc.mk ')' ]; then \
	echo "*** $$f is new; appending to patch.diff" 1>&2; \
	echo New file: "$$f" >>patch.diff; \
	echo diff -u $(DISTDIR)/"$$f" "$$f" >>patch.diff; \
	echo '--- $(DISTDIR)/'"$$f"'	Thu Jan 01 00:00:00 1970' >>patch.diff; \
	diff -u /dev/null "$$f" | tail +2 >>patch.diff; \
	fi; done
	@test -n patch.diff || echo "*** No differences found" 1>&2
	@rm -rf $(DISTDIR) patch.diffpart

diff: patch-check-date patch.diff

patch:
	@test -r patch.diff || (echo "***" 1>&2; \
	echo "*** No 'patch.diff' file found!  Did you remember to" 1>&2; \
	echo "*** run 'make diff'?" 1>&2; \
	echo "***" 1>&2; false)
	@x=`grep "^New file:" patch.diff | head -n 1 | sed 's/New file: //'`; \
	if test -n "$$x" -a -f "$$x"; then \
	echo "*** Note: File '$$x' found in current directory;" 1>&2; \
	echo "*** not applying new files portion of patch.diff." 1>&2; \
	echo "awk '/^New file:/ { exit } { print }' <patch.diff | patch -p0"; \
	awk '/^New file:/ { exit } { print }' <patch.diff | patch -p0; \
	else echo 'patch -p0 <patch.diff'; \
	patch -p0 <patch.diff; \
	fi


# Always build something
always:
	@:

# These are phony targets
.PHONY: all always clean realclean distclean dist distdir tarball handin \
	diff patch patch-extract-tarball patch-check-date patch-make-diff

# Eliminate default suffix rules
.SUFFIXES:

# Delete target files if there is an error (or make is interrupted)
.DELETE_ON_ERROR:

# no intermediate .o files should be deleted
.PRECIOUS: %.o $(OBJDIR)/%.o $(OBJDIR)/%-bootsector
